/*
  File: QueueListCDT.c
  --------------------
  Contains the implementation of the queue abstract data type using a
  simply linked list.
*/

#include <stdio.h>
#include <stdlib.h>


#include "QueueADT.h"

/*
  Definition of the simply linked list node structure
*/

typedef struct ListNode {
  void* obj;
  struct ListNode* next;
} ListNode;



/*
  The queue concrete data type: it consists of a pointer to the head
  and tail of the simply linked list.
*/

typedef struct QueueCDT {
  ListNode* head;
  ListNode* tail;
} QueueCDT;


/*
  Function: NewQueue
  Usage: q = NewQueue();
  ----------------------
  Creates a new empty queue using dynamic memory allocation. If the
  dynamic memory allocation is not successful NULL is returned.
  Head and tail are initialized to NULL.
*/

QueueADT NewQueue(void)
{
  QueueADT q = (QueueADT)malloc(sizeof(QueueCDT));
  if ( q == NULL ) { return NULL; }
  q->head = q->tail = NULL;
  return q;
}



/*
  Function: FreeQueue
  Usage: FreeQueue(q);
  --------------------
  Frees the memory associated with the queue q. If the queue is not
  empty all its elements are deleted using the DeQueue function.
  If q is NULL the program terminates with exit code 1.
*/

void FreeQueue(QueueADT q)
{
  if ( q == NULL ) { exit(1); }
  while (q->head != NULL) { DeQueue(q); }
  free(q);
}



/*
  Function: Enqueue
  Usage: Enqueue(q, (void*)(&item));
  ----------------------------------
  Inserts the element "item" at the end of the queue q.
  The new node of the linked list is created using dynamic memory
  allocation. If the memory allocation fails the program terminates
  with exit code 1.
  If q is NULL the program terminates with exit code 1.
*/

void Enqueue(QueueADT q, void* obj)
{
  if ( q == NULL ) { exit(1); }
  
  ListNode* node = (ListNode*)malloc(sizeof(ListNode));
  if ( node == NULL ) { exit(1); }

  node->obj = obj;
  node->next = NULL;
  if ( q->head == NULL ) {
    q->head = node;
  }
  if ( q->tail != NULL ) {
    q->tail->next = node;
  }
  q->tail = node;
}


/*
  Function: DeQueue
  Usage: p = DeQueue(q);
  ----------------------
  Removes the element at the front of the queue q. The corresponding
  data is returned, and the linked list's node memory is freed.
  If q is NULL the program terminates with exit code 1.
 */

void* DeQueue(QueueADT q)
{
  if ( q == NULL ) { exit(1); }

  void* obj = q->head->obj;
  ListNode* head = q->head;
  q->head = q->head->next;
  free(head);
  return obj;
}


/*
  Function: QueueLength
  Usage: len = QueueLength(q);
  ----------------------------
  Returns the length of the queue q, i.e., the number of elements in q.
  If q is NULL the program terminates with exit code 1.
  This function takes time proportional to the length of the queue.
 */

int QueueLength(QueueADT q)
{
  if ( q == NULL ) { exit(1); }

  int len = 0;

  ListNode* cur = q->head;
  while ( cur != NULL ) {
    ++len;
    cur = cur->next;
  }

  return len;
}



/*
  Function: PrettyPrintQueue
  Usage: PrettyPrintQueue(q);
  ---------------------------
  Print, in a nice way, the length of the queue and the queue elements.
  The queue elements are printed using a printer-function passed via a
  function pointer.
  If q is NULL the program terminates with exit code 1.
 */

void PrettyPrintQueue(QueueADT q, void (*print_element)(void*))
{
  if ( q == NULL ) { exit(1); }

  printf("Queue length: %d\n", QueueLength(q));
  printf("Queue elements:");
  ListNode *cur = q->head;
  while ( cur != NULL ) {
    (*print_element)(cur->obj);
    cur = cur->next;
  }
  printf("\n");
}
